structure DeflateBits :> DeflateBits =
struct

	type byte = Word8.word
	
	(* power of 2 *)
	
	fun pow2 0 = 1 
	  | pow2 1 = 2
	  | pow2 2 = 4
	  | pow2 3 = 8
	  | pow2 4 = 16
	  | pow2 5 = 32
	  | pow2 6 = 64
	  | pow2 7 = 128
	  | pow2 8 = 256 (* is that precomputation useful? *)
	  | pow2 n =
	  		if Int.rem (n, 2) <> 0 then 
				2 * pow2 (n - 1)
			else 
				let val p = pow2 (Int.quot (n, 2)) in p * p end

	(* byte conversion *)
	
	val byte_and = Word8.andb
	fun shiftr (byte, nbits) = Word8.>> (byte, Word.fromInt nbits)
	fun lower_mask nbits = shiftr (0w255, 8 - nbits)
	
	val int_to_byte = Word8.fromInt
	
	fun byte_eq (x,y) = x = y
	
	val byte_to_char = chr o Word8.toInt
	val char_to_byte = int_to_byte o ord
	
	fun is_byte_zero (b:byte) = (b = 0w0)
	
	fun byte_half b = Word8.div (b, 0w2)
	
	(* bool <-> int *)
	
	fun bit_to_int true = 1
	  | bit_to_int false = 0

	fun bit_to_char true = #"1"
	  | bit_to_char false = #"0"

	fun least_bit c = byte_and (c, 0w1) = 0w1

	fun is_even c = Int.rem (c, 2) = 0
	fun is_odd c = not (is_even c)
	
	fun byte_bit bitpos byte = least_bit (shiftr (byte, bitpos))
	
	fun byte_sub (fro,to) byte = Word8.toInt (byte_and (shiftr (byte, fro), lower_mask (to - fro + 1)))

	(* int <-> bool list *)
	
	fun explode_fixed_lsb_first 0 _ = []
	  | explode_fixed_lsb_first len code = 
	  		is_odd code :: explode_fixed_lsb_first (len - 1) (Int.quot (code, 2))

	fun explode_fixed_msb_first len code = rev (explode_fixed_lsb_first len code)

	fun explode_byte_lsb_first n b = explode_fixed_lsb_first n (Word8.toInt b)
		
	fun to_int_lsb_first bits =
		let
			fun f (b, result) = (2 * result) + (bit_to_int b)
		in
			foldr f 0 bits
		end

	fun to_int_msb_first bits = to_int_lsb_first (rev bits)
	
	fun to_string bits = implode (map bit_to_char bits)

	(* self tests *)
	val tests = [
		is_even 2 = true,
		is_even 3 = false,
		is_even 116 = true,
		is_even 117 = false,
		least_bit (int_to_byte 0) = false,
		least_bit (int_to_byte 1) = true,
		least_bit (int_to_byte 16) = false,
		least_bit (int_to_byte 7) = true,
		byte_bit 7 (int_to_byte 128) = true,
		byte_bit 0 (int_to_byte 1) = true,
		bit_to_int true = 1,
		bit_to_int false = 0,
		explode_fixed_msb_first 6 15 = [false,false,true,true,true,true],
		to_int_lsb_first [true,true,false] = 3,
		to_string [false,true,true] = "011",
		to_string [] = ""
		]
end
