structure DeflateHuffman :> DeflateHuffman =
struct

	datatype tree = EMPTY | LEAF of int | NODE of tree * tree

	exception Insert_at_used_location
	exception No_step
	
	val none = EMPTY
	
	fun insert (EMPTY) [] code = LEAF code
	  | insert (NODE (zero,one)) (false::es) code = 
	  		NODE (insert zero es code, one)
	  | insert (NODE (zero,one)) (true::es) code =
  			NODE (zero, insert one es code)
	  | insert (EMPTY) (false::es) code =
  			NODE (insert EMPTY es code, EMPTY)
	  | insert (EMPTY) (true::es) code =
  			NODE (EMPTY, insert EMPTY es code)
	  | insert _ es code = raise Insert_at_used_location

	fun force _ [] code = LEAF code
	  | force (NODE (zero,one)) (false::es) code = 
	  		NODE (force zero es code, one)
	  | force (NODE (zero,one)) (true::es) code =
	  		NODE (zero, force one es code)
	  | force _ (false::es) code =
  			NODE (force EMPTY es code, EMPTY)
	  | force _ (true::es) code =
  			NODE (EMPTY, force EMPTY es code)

	fun at (NODE (zero, _)) (false::es) = at zero es
	  | at (NODE (_, one)) (true::es) = at one es
	  | at tree [] = (tree, [])
	  | at _ es = (EMPTY, es)

	fun code tree es = 
		let 
			fun impl (LEAF c, _) = SOME c
			  | impl _ = NONE 
		in 
			impl (at tree es)
		end

	fun has_step (NODE _) = true
	  | has_step _ = false
	  
	fun step (NODE (zero,_)) false = zero
	  | step (NODE (_, one)) true = one
	  | step _ _ = raise No_step
	   
	fun has_length _ 0 = true
	  | has_length (NODE (zero, one)) len = has_length one (len - 1) orelse has_length zero (len - 1)
	  | has_length _ _ = false
	  		
	fun collapse tree =
		let
			fun sorted ((_, c1),(_, c2)) = c1 <= c2
			fun impl prefix (LEAF c) = [(rev prefix, c)]
			  | impl prefix (EMPTY) = []
			  | impl prefix (NODE (zero, one)) = 
			  		(impl (false::prefix) zero) @ (impl (true::prefix) one)
		in
			DeflateList.sort sorted (impl [] tree)
		end 

	(* for collapse *)
	fun stringify (es, c) = (DeflateBits.to_string es, c)
	
	fun collapse_string tree = map stringify (collapse tree)
	
	(* like collapse_string with ellipsis of successive items *)
	fun collapse_summarised hutree =
		let
			fun succeeding ((p1,c1),(p2,c2)) = length p1 = length p2
				andalso DeflateBits.to_int_msb_first p1 + 1 = DeflateBits.to_int_msb_first p2
				andalso c1 + 1 = c2
			fun f (p, (_,[])) = (p, [p])
			  | f (p, (prev,ps)) =
			  	if succeeding (prev, p) then (p, ps)
				else (p, p :: prev :: ps)
			val input = collapse hutree
			val (_, tmp) = foldl f (([],0),[]) input
			val result = rev (List.last input :: tmp) (* add last item *)
		in
			map stringify result
		end
		
	(* self test *)
	
	val tests = 
		let val tests_tree = NODE (LEAF 10, NODE (NODE (LEAF 20, NODE (LEAF 40, EMPTY)), NODE (LEAF 30, EMPTY)))
		in [
		insert EMPTY [true] 10 = NODE (EMPTY, LEAF 10),
		insert EMPTY [false] 10 = NODE (LEAF 10, EMPTY),
		insert EMPTY [true, true] 10 = NODE (EMPTY, NODE (EMPTY, LEAF 10)),
		insert EMPTY [true, false] 10 = NODE (EMPTY, NODE (LEAF 10, EMPTY)),
		insert (NODE (EMPTY, LEAF 10)) [false] 20 = NODE (LEAF 20, LEAF 10),
		insert (NODE (EMPTY, LEAF 10)) [false,true] 20 = NODE (NODE (EMPTY, LEAF 20), LEAF 10),
		code tests_tree [false] = SOME 10,
		code tests_tree [true, false, false] = SOME 20,
		code tests_tree [true, false, true, false] = SOME 40,
		code tests_tree [true, false, true, true] = NONE,
		code tests_tree [true, true, false] = SOME 30,
		code tests_tree [true, true, true] = NONE,
		has_step (NODE (EMPTY,EMPTY)),
		has_step (LEAF 0) = false,
		has_step (EMPTY) = false,
		step (NODE (LEAF 0, EMPTY)) false = LEAF 0,
		step (NODE (EMPTY, LEAF 1)) true = LEAF 1,
		has_length tests_tree 1,
		has_length tests_tree 2,
		has_length tests_tree 3,
		has_length tests_tree 4,
		has_length tests_tree 5 = false,
		collapse EMPTY = [],
		collapse (NODE (LEAF 1, EMPTY)) = [([false],1)],
		collapse (NODE (NODE (LEAF 1,LEAF 2), LEAF 3)) = 
			[([false,false],1),([false,true],2),([true],3)]
		]
		end
		
end
